@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 /*
- *  Copyright (C) 2004 Hiroyuki Ikezoe
- *  Copyright (C) 2004 Takuro Ashie
+ *  Copyright (C) Hiroyuki Ikezoe <poincare@ikezoe.net>
+ *  Copyright (C) 2004 Takuro Ashie <ashie@homa.ne.jp>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -32,6 +32,11 @@
   #include <config.h>
 #endif
 
+#include <stdio.h>
+#include <unistd.h>
+#include <sys/types.h>
+#include <sys/wait.h>
+
 #include <scim.h>
 #include "scim_anthy_imengine.h"
 #include "scim_anthy_prefs.h"
@@ -71,14 +76,16 @@
 #define SCIM_PROP_PERIOD_STYLE_JAPANESE     "/IMEngine/Anthy/PeriodType/Japanese"
 #define SCIM_PROP_PERIOD_STYLE_WIDE_LATIN   "/IMEngine/Anthy/PeriodType/WideRatin"
 #define SCIM_PROP_PERIOD_STYLE_LATIN        "/IMEngine/Anthy/PeriodType/Ratin"
+#define SCIM_PROP_PERIOD_STYLE_WIDE_LATIN_JAPANESE   "/IMEngine/Anthy/PeriodType/WideRatin_Japanese"
+
+#define SCIM_PROP_DICT                      "/IMEngine/Anthy/Dictionary"
+#define SCIM_PROP_DICT_ADD_WORD             "/IMEngine/Anthy/Dictionary/AddWord"
+#define SCIM_PROP_DICT_LAUNCH_ADMIN_TOOL    "/IMEngine/Anthy/Dictionary/LaunchAdminTool"
 
 #ifndef SCIM_ANTHY_ICON_FILE
     #define SCIM_ANTHY_ICON_FILE           (SCIM_ICONDIR"/scim-anthy.png")
 #endif
 
-// first = name, second = lang
-static KeyEvent __anthy_on_key;
-
 static ConfigPointer _scim_config (0);
 
 extern "C" {
@@ -129,13 +136,19 @@ AnthyFactory::AnthyFactory (const String &lang,
                             const String &uuid,
                             const ConfigPointer &config)
     : m_uuid (uuid),
-      m_config (0),
+      m_config (config),
       m_typing_method (SCIM_ANTHY_CONFIG_TYPING_METHOD_DEFAULT),
       m_period_style (SCIM_ANTHY_CONFIG_PERIOD_STYLE_DEFAULT),
       m_auto_convert (SCIM_ANTHY_CONFIG_AUTO_CONVERT_ON_PERIOD_DEFAULT),
+      m_close_cand_win_on_select (SCIM_ANTHY_CONFIG_CLOSE_CAND_WIN_ON_SELECT),
+      m_dict_admin_command (SCIM_ANTHY_CONFIG_DICT_ADMIN_COMMAND_DEFAULT),
+      m_add_word_command (SCIM_ANTHY_CONFIG_ADD_WORD_COMMAND_DEFAULT),
       m_show_input_mode_label (SCIM_ANTHY_CONFIG_SHOW_INPUT_MODE_LABEL_DEFAULT),
       m_show_typing_method_label (SCIM_ANTHY_CONFIG_SHOW_TYPING_METHOD_LABEL_DEFAULT),
-      m_show_period_style_label (SCIM_ANTHY_CONFIG_SHOW_PERIOD_STYLE_LABEL_DEFAULT)
+      m_show_period_style_label (SCIM_ANTHY_CONFIG_SHOW_PERIOD_STYLE_LABEL_DEFAULT),
+      m_show_dict_label (SCIM_ANTHY_CONFIG_SHOW_DICT_LABEL_DEFAULT),
+      m_show_dict_admin_label (SCIM_ANTHY_CONFIG_SHOW_DICT_ADMIN_LABEL_DEFAULT),
+      m_show_add_word_label (SCIM_ANTHY_CONFIG_SHOW_ADD_WORD_LABEL_DEFAULT)
 {
     SCIM_DEBUG_IMENGINE(1) << "Create Anthy Factory :\n";
     SCIM_DEBUG_IMENGINE(1) << "  Lang : " << lang << "\n";
@@ -148,7 +161,8 @@ AnthyFactory::AnthyFactory (const String &lang,
         return;
 
     /* config */
-    reload_config (config);
+    reload_config (m_config);
+    m_reload_signal_connection = m_config->signal_connect_reload (slot (this, &AnthyFactory::reload_config));
 }
 
 AnthyFactory::~AnthyFactory ()
@@ -209,7 +223,7 @@ AnthyInstance::AnthyInstance (AnthyFactory   *factory,
 {
     SCIM_DEBUG_IMENGINE(1) << "Create Anthy Instance : ";
 
-    // set input mode
+    // set typing method
     if (factory->m_typing_method == "Kana")
         m_preedit.set_typing_method (METHOD_KANA);
     else if (factory->m_typing_method == "Roma")
@@ -217,15 +231,23 @@ AnthyInstance::AnthyInstance (AnthyFactory   *factory,
     else
         m_preedit.set_typing_method (METHOD_ROMAKANA);
 
-    // set typing method
-    if (factory->m_period_style == "WideLatin")
+    // set period style
+    if (factory->m_period_style == "WideLatin") {
+        m_preedit.set_comma_style (COMMA_WIDE_LATIN);
         m_preedit.set_period_style (PERIOD_WIDE_LATIN);
-    else if (factory->m_period_style == "Latin")
+    } else if (factory->m_period_style == "Latin") {
+        m_preedit.set_comma_style (COMMA_LATIN);
         m_preedit.set_period_style (PERIOD_LATIN);
-    else if (factory->m_period_style == "Japanese")
+    } else if (factory->m_period_style == "Japanese") {
+        m_preedit.set_comma_style (COMMA_JAPANESE);
         m_preedit.set_period_style (PERIOD_JAPANESE);
-    else
+    } else if (factory->m_period_style == "WideLatin_Japanese") {
+        m_preedit.set_comma_style (COMMA_WIDE_LATIN);
         m_preedit.set_period_style (PERIOD_JAPANESE);
+    } else {
+        m_preedit.set_comma_style (COMMA_JAPANESE);
+        m_preedit.set_period_style (PERIOD_JAPANESE);
+    }
 
     // set space type
     if (factory->m_space_type == "Half")
@@ -242,19 +264,9 @@ AnthyInstance::AnthyInstance (AnthyFactory   *factory,
 void
 AnthyFactory::reload_config (const ConfigPointer &config)
 {
-    m_reload_signal_connection.disconnect ();
-
     if (config) {
         String str;
 
-#if 1 // FIXME!
-        String on_key
-            = config->read (SCIM_ANTHY_CONFIG_ON_KEY,
-                            String ("Shift+space"));
-        if (!scim_string_to_key (__anthy_on_key, on_key))
-            __anthy_on_key = KeyEvent (SCIM_KEY_space, SCIM_KEY_ShiftMask);
-#endif
-
         m_typing_method
             = config->read (SCIM_ANTHY_CONFIG_TYPING_METHOD,
                             m_typing_method);
@@ -267,6 +279,15 @@ AnthyFactory::reload_config (const ConfigPointer &config)
         m_auto_convert
             = config->read (SCIM_ANTHY_CONFIG_AUTO_CONVERT_ON_PERIOD,
                             m_auto_convert);
+        m_close_cand_win_on_select
+            = config->read (SCIM_ANTHY_CONFIG_CLOSE_CAND_WIN_ON_SELECT,
+                            m_close_cand_win_on_select);
+        m_dict_admin_command
+            = config->read (SCIM_ANTHY_CONFIG_DICT_ADMIN_COMMAND,
+                            m_dict_admin_command);
+        m_add_word_command
+            = config->read (SCIM_ANTHY_CONFIG_ADD_WORD_COMMAND,
+                            m_add_word_command);
         m_show_input_mode_label
             = config->read (SCIM_ANTHY_CONFIG_SHOW_INPUT_MODE_LABEL,
                             m_show_input_mode_label);
@@ -276,6 +297,15 @@ AnthyFactory::reload_config (const ConfigPointer &config)
         m_show_period_style_label
             = config->read (SCIM_ANTHY_CONFIG_SHOW_PERIOD_STYLE_LABEL,
                             m_show_period_style_label);
+        m_show_dict_label
+            = config->read (SCIM_ANTHY_CONFIG_SHOW_DICT_LABEL,
+                            m_show_dict_label);
+        m_show_dict_admin_label
+            = config->read (SCIM_ANTHY_CONFIG_SHOW_DICT_ADMIN_LABEL,
+                            m_show_dict_admin_label);
+        m_show_add_word_label
+            = config->read (SCIM_ANTHY_CONFIG_SHOW_ADD_WORD_LABEL,
+                            m_show_add_word_label);
 
         // edit keys
         str = config->read (String (SCIM_ANTHY_CONFIG_COMMIT_KEY),
@@ -365,6 +395,46 @@ AnthyFactory::reload_config (const ConfigPointer &config)
                             String (SCIM_ANTHY_CONFIG_CANDIDATES_PAGE_DOWN_KEY_DEFAULT));
         scim_string_to_key_list (m_candidates_page_down_keys, str);
 
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_1_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_1_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[0], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_2_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_2_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[1], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_3_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_3_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[2], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_4_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_4_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[3], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_5_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_5_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[4], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_6_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_6_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[5], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_7_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_7_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[6], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_8_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_8_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[7], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_9_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_9_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[8], str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_10_KEY),
+                            String (SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_10_KEY_DEFAULT));
+        scim_string_to_key_list (m_select_candidate_keys[9], str);
+
         // convert keys
         str = config->read (String (SCIM_ANTHY_CONFIG_CONV_TO_HIRAGANA_KEY),
                             String (SCIM_ANTHY_CONFIG_CONV_TO_HIRAGANA_KEY_DEFAULT));
@@ -398,10 +468,16 @@ AnthyFactory::reload_config (const ConfigPointer &config)
         str = config->read (String (SCIM_ANTHY_CONFIG_CIRCLE_KANA_MODE_KEY),
                             String (SCIM_ANTHY_CONFIG_CIRCLE_KANA_MODE_KEY_DEFAULT));
         scim_string_to_key_list (m_circle_kana_mode_keys, str);
-    }
 
-    m_config = config;
-    m_reload_signal_connection = m_config->signal_connect_reload (slot (this, &AnthyFactory::reload_config));
+        // dict keys
+        str = config->read (String (SCIM_ANTHY_CONFIG_DICT_ADMIN_KEY),
+                            String (SCIM_ANTHY_CONFIG_DICT_ADMIN_KEY_DEFAULT));
+        scim_string_to_key_list (m_dict_admin_keys, str);
+
+        str = config->read (String (SCIM_ANTHY_CONFIG_ADD_WORD_KEY),
+                            String (SCIM_ANTHY_CONFIG_ADD_WORD_KEY_DEFAULT));
+        scim_string_to_key_list (m_add_word_keys, str);
+    }
 }
 
 AnthyInstance::~AnthyInstance ()
@@ -499,6 +575,46 @@ AnthyInstance::process_key_event_lookup_keybind (const KeyEvent& key)
         action_candidates_page_down ())
         return true;
 
+    if (match_key_event (m_factory->m_select_candidate_keys[0], key) &&
+        action_select_candidate_1 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[1], key) &&
+        action_select_candidate_2 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[2], key) &&
+        action_select_candidate_3 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[3], key) &&
+        action_select_candidate_4 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[4], key) &&
+        action_select_candidate_5 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[5], key) &&
+        action_select_candidate_6 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[6], key) &&
+        action_select_candidate_7 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[7], key) &&
+        action_select_candidate_8 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[8], key) &&
+        action_select_candidate_9 ())
+        return true;
+
+    if (match_key_event (m_factory->m_select_candidate_keys[9], key) &&
+        action_select_candidate_10 ())
+        return true;
+
     // convert
     if (match_key_event (m_factory->m_conv_to_hiragana_keys, key) &&
         action_convert_to_hiragana ())
@@ -533,6 +649,15 @@ AnthyInstance::process_key_event_lookup_keybind (const KeyEvent& key)
         action_toggle_wide_latin_mode ())
         return true;
 
+    // dict
+    if (match_key_event (m_factory->m_dict_admin_keys, key) &&
+        action_launch_dict_admin_tool ())
+        return true;
+
+    if (match_key_event (m_factory->m_add_word_keys, key) &&
+        action_add_word ())
+        return true;
+
     return false;
 }
 
@@ -551,26 +676,6 @@ AnthyInstance::process_key_event_with_preedit (const KeyEvent& key)
 bool
 AnthyInstance::process_key_event_with_candidate (const KeyEvent &key)
 {
-    switch (key.code) {
-    case SCIM_KEY_1: 
-    case SCIM_KEY_2: 
-    case SCIM_KEY_3: 
-    case SCIM_KEY_4: 
-    case SCIM_KEY_5: 
-    case SCIM_KEY_6: 
-    case SCIM_KEY_7: 
-    case SCIM_KEY_8: 
-    case SCIM_KEY_9: 
-        select_candidate (key.code - SCIM_KEY_1);
-        return true;
-    case SCIM_KEY_0:
-        select_candidate (9);
-        return true;
-
-    default:
-        break;
-    }
-
     return process_remaining_key_event (key);
 }
 
@@ -586,8 +691,8 @@ AnthyInstance::process_remaining_key_event (const KeyEvent &key)
     {
         return false;
     }
-
-    if (isprint(key.code)) {
+   
+    if (isprint(key.get_ascii_code ())) {
         // commit old conversion string before update preedit string
         if (m_preedit.is_converting ())
             action_commit ();
@@ -615,22 +720,9 @@ AnthyInstance::process_key_event (const KeyEvent& key)
     SCIM_DEBUG_IMENGINE(2) << "process_key_event.\n";
     KeyEvent newkey;
 
-    newkey.code = key.code;
-    newkey.mask = key.mask & (SCIM_KEY_ShiftMask | SCIM_KEY_ControlMask | SCIM_KEY_AltMask | SCIM_KEY_ReleaseMask);
-
-#if 0
-    // change input mode
-    if (__anthy_on_key.is_key_press ()) {
-        trigger_property (SCIM_PROP_INPUT_MODE);
-        return true;
-    }
-#endif
-
-    m_prev_key = key;
-
     // ignore key release.
     if (key.is_key_release ())
-        return false;
+        return true;
 
     // ignore modifier keys
     if (key.code == SCIM_KEY_Shift_L || key.code == SCIM_KEY_Shift_R ||
@@ -642,6 +734,9 @@ AnthyInstance::process_key_event (const KeyEvent& key)
     if (process_key_event_lookup_keybind (key))
         return true;
 
+    if (m_preedit.get_input_mode () == MODE_LATIN)
+        return false;
+
     // process hard coded keys
     if (is_selecting_candidates ())
         return process_key_event_with_candidate (key);
@@ -659,11 +754,11 @@ AnthyInstance::move_preedit_caret (unsigned int pos)
 }
 
 void
-AnthyInstance::select_candidate (unsigned int item)
+AnthyInstance::select_candidate_no_direct (unsigned int item)
 {
     if (!is_selecting_candidates ()) return;
 
-    SCIM_DEBUG_IMENGINE(2) << "select_candidate.\n";
+    SCIM_DEBUG_IMENGINE(2) << "select_candidate_no_direct.\n";
  
     // update lookup table
     m_lookup_table.set_cursor_pos_in_current_page (item);
@@ -676,6 +771,20 @@ AnthyInstance::select_candidate (unsigned int item)
     update_preedit_caret (m_preedit.get_caret_pos ());
 }
 
+void
+AnthyInstance::select_candidate (unsigned int item)
+{
+    SCIM_DEBUG_IMENGINE(2) << "select_candidate.\n";
+
+    select_candidate_no_direct (item);
+
+    if (m_factory->m_close_cand_win_on_select) {
+        m_lookup_table.clear ();
+        hide_lookup_table ();
+        action_select_next_segment();
+    }
+}
+
 void
 AnthyInstance::update_lookup_table_page_size (unsigned int page_size)
 {
@@ -718,6 +827,8 @@ AnthyInstance::reset ()
 
     m_preedit.clear ();
     m_lookup_table.clear ();
+    update_preedit_caret (0);
+    update_preedit_string (utf8_mbstowcs (""));
     hide_lookup_table ();
     hide_preedit_string ();
 }
@@ -743,6 +854,10 @@ void
 AnthyInstance::focus_out ()
 {
     SCIM_DEBUG_IMENGINE(2) << "focus_out.\n";
+    
+    // the following line should be commented out because of 
+    // crashing evolution.
+    //action_commit();
 }
 
 void
@@ -803,6 +918,11 @@ AnthyInstance::install_properties (void)
                              "\xE3\x80\x81\xE3\x80\x82");
             m_properties.push_back (prop);
 
+            prop = Property (SCIM_PROP_PERIOD_STYLE_WIDE_LATIN_JAPANESE,
+                             "\xEF\xBC\x8C\xE3\x80\x82", String (""),
+                             "\xEF\xBC\x8C\xE3\x80\x82");
+            m_properties.push_back (prop);
+
             prop = Property (SCIM_PROP_PERIOD_STYLE_WIDE_LATIN,
                              "\xEF\xBC\x8C\xEF\xBC\x8E", String (""),
                              "\xEF\xBC\x8C\xEF\xBC\x8E");
@@ -812,11 +932,35 @@ AnthyInstance::install_properties (void)
                              ",.", String (""), ",.");
             m_properties.push_back (prop);
         }
+
+        if (m_factory->m_show_dict_label) {
+            prop = Property (SCIM_PROP_DICT,
+                             String(""), //_("Dictionary"),
+                             String (SCIM_ICONDIR "/" "scim-anthy-dict.png"),
+                             _("Dictionary menu"));
+            m_properties.push_back (prop);
+
+            if (m_factory->m_show_dict_admin_label) {
+                prop = Property (SCIM_PROP_DICT_LAUNCH_ADMIN_TOOL,
+                                 _("Edit the dictionary"),
+                                 String (SCIM_ICONDIR "/" "scim-anthy-dict.png"),
+                                 _("Launch the dictionary administration tool."));
+                m_properties.push_back (prop);
+            }
+
+            if (m_factory->m_show_add_word_label) {
+                prop = Property (SCIM_PROP_DICT_ADD_WORD,
+                                 _("Add a word"),
+                                 String (SCIM_ICONDIR "/" "scim-anthy-dict.png"),
+                                 _("Add a word to the dictorinay."));
+                m_properties.push_back (prop);
+            }
+        }
     }
 
     set_input_mode (m_preedit.get_input_mode ());
     set_typing_method (m_preedit.get_typing_method ());
-    set_period_style (m_preedit.get_period_style ());
+    set_period_style (m_preedit.get_period_style (), m_preedit.get_comma_style ());
 
     register_properties (m_properties);
 }
@@ -891,36 +1035,52 @@ AnthyInstance::set_typing_method (TypingMethod method)
 }
 
 void
-AnthyInstance::set_period_style (PeriodStyle style)
+AnthyInstance::set_period_style (PeriodStyle period, CommaStyle comma)
 {
-    const char *label = "";
+    String label;
 
-    switch (style) {
-    case PERIOD_JAPANESE:
-        label = "\xE3\x80\x81\xE3\x80\x82";
+    switch (comma) {
+    case COMMA_JAPANESE:
+        label = "\xE3\x80\x81";
         break;
-    case PERIOD_WIDE_LATIN:
-        label = "\xEF\xBC\x8C\xEF\xBC\x8E";
+    case COMMA_WIDE_LATIN:
+        label = "\xEF\xBC\x8C";
         break;
-    case PERIOD_LATIN:
-        label = ",.";
+    case COMMA_LATIN:
+        label = ",";
         break;
     default:
         break;
     }
 
-    if (label && *label /*&& m_factory->m_show_typing_method_label*/) {
+    switch (period) {
+    case PERIOD_JAPANESE:
+        label += "\xE3\x80\x82";
+        break;
+    case PERIOD_WIDE_LATIN:
+        label += "\xEF\xBC\x8E";
+        break;
+    case PERIOD_LATIN:
+        label += ".";
+        break;
+    default:
+        break;
+    }
+
+    if (label.length () > 0) {
         PropertyList::iterator it = std::find (m_properties.begin (),
                                                m_properties.end (),
                                                SCIM_PROP_PERIOD_STYLE);
         if (it != m_properties.end ()) {
-            it->set_label (label);
+            it->set_label (label.c_str ());
             update_property (*it);
         }
     }
 
-    if (style != m_preedit.get_period_style ())
-        m_preedit.set_period_style (style);
+    if (period != m_preedit.get_period_style ())
+        m_preedit.set_period_style (period);
+    if (comma != m_preedit.get_comma_style ())
+        m_preedit.set_comma_style (comma);
 }
 
 bool
@@ -954,24 +1114,24 @@ AnthyInstance::action_convert (void)
         // show candidates window
         // FIXME!
         m_preedit.setup_lookup_table(m_lookup_table);
-	int page = m_preedit.get_selected_candidate () / m_lookup_table.get_page_size ();
+        int page = m_preedit.get_selected_candidate () / m_lookup_table.get_page_size ();
         int pos  = m_preedit.get_selected_candidate () % m_lookup_table.get_page_size ();
-        if (m_preedit.get_selected_candidate () >= m_lookup_table.number_of_candidates () - 1) {
-	    m_lookup_table.set_cursor_pos (0);
-	    select_candidate (0);
-	} else {
-	    for (int i = 0; i < page; i++)
+        if (m_preedit.get_selected_candidate () >= (int) (m_lookup_table.number_of_candidates () - 1)) {
+            m_lookup_table.set_cursor_pos (0);
+            select_candidate_no_direct (0);
+        } else {
+            for (int i = 0; i < page; i++)
                 m_lookup_table.page_down ();
             if (pos + 1 >= m_lookup_table.get_page_size ()) {
-	        m_lookup_table.page_down ();
-	        select_candidate (0);
+                m_lookup_table.page_down ();
+                select_candidate_no_direct (0);
             } else {
-                select_candidate (pos + 1);
+                select_candidate_no_direct (pos + 1);
             }
-	}
+        }
         show_lookup_table ();
     }
-
+    
     return true;
 }
 
@@ -1294,11 +1454,13 @@ AnthyInstance::action_select_next_candidate (void)
 
     if (!is_selecting_candidates ())
         action_convert ();
-    if (m_lookup_table.get_cursor_pos () == m_lookup_table.number_of_candidates () - 1)
+    if (m_lookup_table.get_cursor_pos () == (int) (m_lookup_table.number_of_candidates () - 1))
+    {
         m_lookup_table.set_cursor_pos (0);
-    else
+    } else {
         m_lookup_table.cursor_down ();
-    select_candidate (m_lookup_table.get_cursor_pos_in_current_page ());
+    }
+    select_candidate_no_direct (m_lookup_table.get_cursor_pos_in_current_page ());
 
     return true;
 }
@@ -1315,7 +1477,7 @@ AnthyInstance::action_select_prev_candidate (void)
         m_lookup_table.set_cursor_pos (m_lookup_table.number_of_candidates () - 1);
     else
         m_lookup_table.cursor_up ();
-    select_candidate (m_lookup_table.get_cursor_pos_in_current_page ());
+    select_candidate_no_direct (m_lookup_table.get_cursor_pos_in_current_page ());
 
     return true;
 }
@@ -1327,7 +1489,7 @@ AnthyInstance::action_candidates_page_up(void)
     if (!is_selecting_candidates ()) return false;
 
     m_lookup_table.page_up ();
-    select_candidate (m_lookup_table.get_cursor_pos_in_current_page ());
+    select_candidate_no_direct (m_lookup_table.get_cursor_pos_in_current_page ());
 
     return true;
 }
@@ -1339,11 +1501,83 @@ AnthyInstance::action_candidates_page_down (void)
     if (!is_selecting_candidates ()) return false;
 
     m_lookup_table.page_down ();
-    select_candidate (m_lookup_table.get_cursor_pos_in_current_page ());
+    select_candidate_no_direct (m_lookup_table.get_cursor_pos_in_current_page ());
 
     return true;
 }
 
+bool
+AnthyInstance::action_select_candidate (unsigned int i)
+{
+    if (!m_preedit.is_converting ()) return false;
+    if (!is_selecting_candidates ()) return false;
+
+    select_candidate (i);
+
+    return true;
+}
+
+bool
+AnthyInstance::action_select_candidate_1 (void)
+{
+    return action_select_candidate (0);
+}
+
+bool
+AnthyInstance::action_select_candidate_2 (void)
+{
+    return action_select_candidate (1);
+}
+
+bool
+AnthyInstance::action_select_candidate_3 (void)
+{
+    return action_select_candidate (2);
+}
+
+bool
+AnthyInstance::action_select_candidate_4 (void)
+{
+    return action_select_candidate (3);
+}
+
+bool
+AnthyInstance::action_select_candidate_5 (void)
+{
+    return action_select_candidate (4);
+}
+
+bool
+AnthyInstance::action_select_candidate_6 (void)
+{
+    return action_select_candidate (5);
+}
+
+bool
+AnthyInstance::action_select_candidate_7 (void)
+{
+    return action_select_candidate (6);
+}
+
+
+bool
+AnthyInstance::action_select_candidate_8 (void)
+{
+    return action_select_candidate (7);
+}
+
+bool
+AnthyInstance::action_select_candidate_9 (void)
+{
+    return action_select_candidate (8);
+}
+
+bool
+AnthyInstance::action_select_candidate_10 (void)
+{
+    return action_select_candidate (9);
+}
+
 bool
 AnthyInstance::action_circle_input_mode (void)
 {
@@ -1453,7 +1687,7 @@ AnthyInstance::action_toggle_wide_latin_mode (void)
 }
 
 bool
-AnthyInstance::convert_kana (SpecialCandidate type)
+AnthyInstance::convert_kana (CandidateType type)
 {
     if (!m_preedit.is_preediting ())
         return false;
@@ -1506,16 +1740,86 @@ AnthyInstance::action_convert_to_wide_latin (void)
     return convert_kana (CANDIDATE_WIDE_LATIN);
 }
 
+static void
+launch_program (const char *command)
+{
+    if (!command) return;
+
+    /* split string */
+    unsigned int len = strlen (command);
+    char tmp[len + 1];
+    strncpy (tmp, command, len);
+    tmp[len] = '\0';
+
+    char *str = tmp;
+    std::vector<char *> array;
+
+    for (unsigned int i = 0; i < len + 1; i++) {
+        if (!tmp[i] || isspace (tmp[i])) {
+            if (*str) {
+                tmp[i] = '\0';
+                array.push_back (str);
+                printf("str: %s\n", str);
+            }
+            str = tmp + i + 1;
+        }
+    }
+
+    if (array.size () <= 0) return;
+    array.push_back (NULL);
+
+    char *args[array.size()];
+    for (unsigned int i = 0; i < array.size (); i++)
+        args[i] = array[i];
+
+
+    /* exec command */
+	pid_t child_pid;
+
+	child_pid = fork();
+	if (child_pid < 0) {
+		perror("fork");
+	} else if (child_pid == 0) {		 /* child process  */
+		pid_t grandchild_pid;
+
+		grandchild_pid = fork();
+		if (grandchild_pid < 0) {
+			perror("fork");
+			_exit(1);
+		} else if (grandchild_pid == 0) { /* grandchild process  */
+			execvp(args[0], args);
+			perror("execvp");
+			_exit(1);
+		} else {
+			_exit(0);
+		}
+	} else {                              /* parent process */
+		int status;
+		waitpid(child_pid, &status, 0);
+	}
+}
+
+bool
+AnthyInstance::action_add_word (void)
+{
+    launch_program (m_factory->m_add_word_command.c_str ());
+
+    return true;
+}
+
+bool
+AnthyInstance::action_launch_dict_admin_tool (void)
+{
+    launch_program (m_factory->m_dict_admin_command.c_str ());
+
+    return true;
+}
+
 #if 0
 void
 AnthyInstance::action_regist_word (void)
 {
 }
-
-void
-AnthyInstance::action_launch_dict_admin (void)
-{
-}
 #endif
 
 void
@@ -1535,16 +1839,25 @@ AnthyInstance::trigger_property (const String &property)
         set_input_mode (MODE_LATIN);
     } else if (property == SCIM_PROP_INPUT_MODE_WIDE_LATIN) {
         set_input_mode (MODE_WIDE_LATIN);
+
     } else if (property == SCIM_PROP_TYPING_METHOD_ROMAKANA) {
         set_typing_method (METHOD_ROMAKANA);
     } else if (property == SCIM_PROP_TYPING_METHOD_KANA) {
         set_typing_method (METHOD_KANA);
+
     } else if (property == SCIM_PROP_PERIOD_STYLE_JAPANESE) {
-        set_period_style (PERIOD_JAPANESE);
+        set_period_style (PERIOD_JAPANESE, COMMA_JAPANESE);
+    } else if (property == SCIM_PROP_PERIOD_STYLE_WIDE_LATIN_JAPANESE) {
+        set_period_style (PERIOD_JAPANESE, COMMA_WIDE_LATIN);
     } else if (property == SCIM_PROP_PERIOD_STYLE_WIDE_LATIN) {
-        set_period_style (PERIOD_WIDE_LATIN);
+        set_period_style (PERIOD_WIDE_LATIN, COMMA_WIDE_LATIN);
     } else if (property == SCIM_PROP_PERIOD_STYLE_LATIN) {
-        set_period_style (PERIOD_LATIN);
+        set_period_style (PERIOD_LATIN, COMMA_LATIN);
+
+    } else if (property == SCIM_PROP_DICT_ADD_WORD) {
+        action_add_word ();
+    } else if (property == SCIM_PROP_DICT_LAUNCH_ADMIN_TOOL) {
+        action_launch_dict_admin_tool ();
     }
 }
 
@@ -1555,11 +1868,10 @@ AnthyInstance::match_key_event (const KeyEventList &keys, const KeyEvent &key) c
 
     for (kit = keys.begin (); kit != keys.end (); ++kit) {
         if (key.code == kit->code && key.mask == kit->mask)
-            if (!(key.mask & SCIM_KEY_ReleaseMask) ||
-                m_prev_key.code == key.code)
-            {
-                return true;
-            }
+             return true;
     }
     return false;
 }
+/*
+vi:ts=4:nowrap:ai:expandtab
+*/
