@@ -128,6 +128,12 @@ struct KeyboardConfigData
     GtkWidget  *button;
 };
 
+struct KeyboardConfigPage
+{
+    const char         *label;
+    KeyboardConfigData *data;
+};
+
 struct ComboConfigData
 {
     const char *label;
@@ -138,10 +144,16 @@ struct ComboConfigData
 static String __config_typing_method            = SCIM_ANTHY_CONFIG_TYPING_METHOD_DEFAULT;
 static String __config_period_style             = SCIM_ANTHY_CONFIG_PERIOD_STYLE_DEFAULT;
 static String __config_space_type               = SCIM_ANTHY_CONFIG_SPACE_TYPE_DEFAULT;
+static String __config_dict_admin_command       = SCIM_ANTHY_CONFIG_DICT_ADMIN_COMMAND_DEFAULT;
+static String __config_add_word_command         = SCIM_ANTHY_CONFIG_ADD_WORD_COMMAND_DEFAULT;
 static bool   __config_auto_convert             = SCIM_ANTHY_CONFIG_AUTO_CONVERT_ON_PERIOD_DEFAULT;
+static bool   __config_close_cand_win_on_select = SCIM_ANTHY_CONFIG_CLOSE_CAND_WIN_ON_SELECT_DEFAULT;
 static bool   __config_show_input_mode_label    = SCIM_ANTHY_CONFIG_SHOW_INPUT_MODE_LABEL_DEFAULT;
 static bool   __config_show_typing_method_label = SCIM_ANTHY_CONFIG_SHOW_TYPING_METHOD_LABEL_DEFAULT;
 static bool   __config_show_period_style_label  = SCIM_ANTHY_CONFIG_SHOW_PERIOD_STYLE_LABEL_DEFAULT;
+static bool   __config_show_dict_label          = SCIM_ANTHY_CONFIG_SHOW_DICT_LABEL_DEFAULT;
+static bool   __config_show_dict_admin_label    = SCIM_ANTHY_CONFIG_SHOW_DICT_ADMIN_LABEL_DEFAULT;
+static bool   __config_show_add_word_label      = SCIM_ANTHY_CONFIG_SHOW_ADD_WORD_LABEL_DEFAULT;
 
 static bool __have_changed    = true;
 
@@ -149,14 +161,19 @@ static GtkWidget    * __widget_typing_method            = 0;
 static GtkWidget    * __widget_period_style             = 0;
 static GtkWidget    * __widget_space_type               = 0;
 static GtkWidget    * __widget_auto_convert             = 0;
+static GtkWidget    * __widget_close_cand_win_on_select  = 0;
+static GtkWidget    * __widget_dict_admin_command       = 0;
+static GtkWidget    * __widget_add_word_command         = 0;
 static GtkWidget    * __widget_show_input_mode_label    = 0;
 static GtkWidget    * __widget_show_typing_method_label = 0;
 static GtkWidget    * __widget_show_period_style_label  = 0;
+static GtkWidget    * __widget_show_dict_label          = 0;
+static GtkWidget    * __widget_show_dict_admin_label    = 0;
+static GtkWidget    * __widget_show_add_word_label      = 0;
 static GtkTooltips  * __widget_tooltips                 = 0;
 
-static KeyboardConfigData __config_keyboards [] =
+static KeyboardConfigData __config_keyboards_common [] =
 {
-    /* edit */
     {
         SCIM_ANTHY_CONFIG_COMMIT_KEY,
         SCIM_ANTHY_CONFIG_COMMIT_KEY_DEFAULT,
@@ -202,8 +219,19 @@ static KeyboardConfigData __config_keyboards [] =
         NULL,
         NULL,
     },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
+};
 
-    /* caret */
+static KeyboardConfigData __config_keyboards_caret [] =
+{
     {
         SCIM_ANTHY_CONFIG_MOVE_CARET_FIRST_KEY,
         SCIM_ANTHY_CONFIG_MOVE_CARET_FIRST_KEY_DEFAULT,
@@ -240,8 +268,19 @@ static KeyboardConfigData __config_keyboards [] =
         NULL,
         NULL,
     },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
+};
 
-    /* segment */
+static KeyboardConfigData __config_keyboards_segments [] =
+{
     {
         SCIM_ANTHY_CONFIG_SELECT_FIRST_SEGMENT_KEY,
         SCIM_ANTHY_CONFIG_SELECT_FIRST_SEGMENT_KEY_DEFAULT,
@@ -314,8 +353,19 @@ static KeyboardConfigData __config_keyboards [] =
         NULL,
         NULL,
     },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
+};
 
-    /* candidates */
+static KeyboardConfigData __config_keyboards_candidates [] =
+{
     {
         SCIM_ANTHY_CONFIG_SELECT_NEXT_CANDIDATE_KEY,
         SCIM_ANTHY_CONFIG_SELECT_NEXT_CANDIDATE_KEY_DEFAULT,
@@ -352,8 +402,122 @@ static KeyboardConfigData __config_keyboards [] =
         NULL,
         NULL,
     },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
+};
 
-    /* convert */
+static KeyboardConfigData __config_keyboards_direct_select_candidate [] =
+{
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_1_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_1_KEY_DEFAULT,
+        N_("1st candidate keys:"),
+        N_("Select keys to select 1st candidate"),
+        N_("The key events to select 1st candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_2_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_2_KEY_DEFAULT,
+        N_("2nd candidate keys:"),
+        N_("Select keys to select 2nd candidate"),
+        N_("The key events to select 2nd candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_3_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_3_KEY_DEFAULT,
+        N_("3rd candidate keys:"),
+        N_("Select keys to select 3rd candidate"),
+        N_("The key events to select 3rd candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_4_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_4_KEY_DEFAULT,
+        N_("4th candidate keys:"),
+        N_("Select keys to select 4th candidate"),
+        N_("The key events to select 4th candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_5_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_5_KEY_DEFAULT,
+        N_("5th candidate keys:"),
+        N_("Select keys to select 5th candidate"),
+        N_("The key events to select 5th candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_6_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_6_KEY_DEFAULT,
+        N_("6th candidate keys:"),
+        N_("Select keys to select 6th candidate"),
+        N_("The key events to select 6th candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_7_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_7_KEY_DEFAULT,
+        N_("7th candidate keys:"),
+        N_("Select keys to select 7th candidate"),
+        N_("The key events to select 7th candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_8_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_8_KEY_DEFAULT,
+        N_("8th candidate keys:"),
+        N_("Select keys to select 8th candidate"),
+        N_("The key events to select 8th candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_9_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_9_KEY_DEFAULT,
+        N_("9th candidate keys:"),
+        N_("Select keys to select 9th candidate"),
+        N_("The key events to select 9th candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_10_KEY,
+        SCIM_ANTHY_CONFIG_SELECT_CANDIDATE_10_KEY_DEFAULT,
+        N_("10th candidate keys:"),
+        N_("Select keys to select 10th candidate"),
+        N_("The key events to select 10th candidate. "),
+        NULL,
+        NULL,
+    },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
+};
+
+static KeyboardConfigData __config_keyboards_converting [] =
+{
     {
         SCIM_ANTHY_CONFIG_CONV_TO_HIRAGANA_KEY,
         SCIM_ANTHY_CONFIG_CONV_TO_HIRAGANA_KEY_DEFAULT,
@@ -399,8 +563,19 @@ static KeyboardConfigData __config_keyboards [] =
         NULL,
         NULL,
     },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
+};
 
-    /* mode */
+static KeyboardConfigData __config_keyboards_mode [] =
+{
     {
         SCIM_ANTHY_CONFIG_CIRCLE_KANA_MODE_KEY,
         SCIM_ANTHY_CONFIG_CIRCLE_KANA_MODE_KEY_DEFAULT,
@@ -438,28 +613,50 @@ static KeyboardConfigData __config_keyboards [] =
         NULL,
     },
 };
-/* FIXME! */
-#if 1
-static unsigned int key_nums[] =
+
+static KeyboardConfigData __config_keyboards_dict [] =
 {
-    5, // edit
-    4, // caret
-    8, // segment
-    4, // candidates
-    5, // convert
-    3, // mode
+    {
+        SCIM_ANTHY_CONFIG_DICT_ADMIN_KEY,
+        SCIM_ANTHY_CONFIG_DICT_ADMIN_KEY_DEFAULT,
+        N_("Edit dictionary keys:"),
+        N_("Select edit dictionary keys"),
+        N_("The key events to launch dictionary administration tool. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_ANTHY_CONFIG_ADD_WORD_KEY,
+        SCIM_ANTHY_CONFIG_ADD_WORD_KEY_DEFAULT,
+        N_("Add a word keys:"),
+        N_("Select add a word keys"),
+        N_("The key events to launch the tool to add a word. "),
+        NULL,
+        NULL,
+    },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
 };
-static char *key_conf_page_labels[] =
+
+static struct KeyboardConfigPage __key_conf_pages[] =
 {
-    N_("Common keys"),
-    N_("Caret keys"),
-    N_("Segments keys"),
-    N_("Candidates keys"),
-    N_("Converting keys"),
-    N_("Mode keys"),
+    {N_("Common keys"),     __config_keyboards_common},
+    {N_("Mode keys"),       __config_keyboards_mode},
+    {N_("Caret keys"),      __config_keyboards_caret},
+    {N_("Segments keys"),   __config_keyboards_segments},
+    {N_("Candidates keys"), __config_keyboards_candidates},
+    {N_("Candidates keys (Direct select)"), __config_keyboards_direct_select_candidate},
+    {N_("Converting keys"), __config_keyboards_converting},
+    {N_("Dictionary keys"), __config_keyboards_dict},
 };
-static unsigned int key_conf_pages = sizeof (key_nums) / sizeof (unsigned int);
-#endif
+static unsigned int __key_conf_pages_num = sizeof (__key_conf_pages) / sizeof (KeyboardConfigPage);
 
 static ComboConfigData typing_methods[] =
 {
@@ -470,9 +667,10 @@ static ComboConfigData typing_methods[] =
 
 static ComboConfigData period_styles[] =
 {
-    {"\xE3\x80\x81\xE3\x80\x82", "Japanese",},
+    {"\xE3\x80\x81\xE3\x80\x82", "Japanese"},
+    {"\xEF\xBC\x8C\xE3\x80\x82", "WideLatin_Japanese"},
     {"\xEF\xBC\x8C\xEF\xBC\x8E", "WideLatin"},
-    {", .",                      "Latin"},
+    {",.",                       "Latin"},
     {NULL, NULL},
 };
 
@@ -484,15 +682,17 @@ static ComboConfigData space_types[] =
 };
 
 
-static void on_default_editable_changed      (GtkEditable     *editable,
-                                              gpointer         user_data);
-static void on_default_toggle_button_toggled (GtkToggleButton *togglebutton,
-                                              gpointer         user_data);
-static void on_default_key_selection_clicked (GtkButton       *button,
-                                              gpointer         user_data);
-static void on_default_combo_changed         (GtkEditable     *editable,
-                                              gpointer         user_data);
-void setup_widget_value ();
+static void on_default_editable_changed       (GtkEditable     *editable,
+                                               gpointer         user_data);
+static void on_default_toggle_button_toggled  (GtkToggleButton *togglebutton,
+                                               gpointer         user_data);
+static void on_default_key_selection_clicked  (GtkButton       *button,
+                                               gpointer         user_data);
+static void on_default_combo_changed          (GtkEditable     *editable,
+                                               gpointer         user_data);
+static void on_dict_menu_label_toggled        (GtkToggleButton *togglebutton,
+                                               gpointer         user_data);
+static void setup_widget_value ();
 
 
 static GtkWidget *
@@ -528,33 +728,33 @@ create_combo_widget (const char *label_text, GtkWidget **widget,
 static GtkWidget *
 create_options_page ()
 {
-    GtkWidget *vbox, *hbox;
+    GtkWidget *vbox, *widget;
 
     vbox = gtk_vbox_new (FALSE, 0);
     gtk_widget_show (vbox);
 
     /* typing method */
-    hbox = create_combo_widget (_("Typing method: "),
-                                &__widget_typing_method,
-                                (gpointer) &__config_typing_method,
-                                (gpointer) &typing_methods);
-    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
+    widget = create_combo_widget (_("Typing method: "),
+                                  &__widget_typing_method,
+                                  (gpointer) &__config_typing_method,
+                                  (gpointer) &typing_methods);
+    gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 4);
 
-    /* comma and period style */
-    hbox = create_combo_widget (_("Style of comma and period: "),
-                                &__widget_period_style,
-                                (gpointer) &__config_period_style,
-                                (gpointer) &period_styles);
+    /* period style */
+    widget = create_combo_widget (_("Style of comma and period: "),
+                                  &__widget_period_style,
+                                  (gpointer) &__config_period_style,
+                                  (gpointer) &period_styles);
     gtk_widget_set_size_request (__widget_period_style, 100, -1);
-    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
+    gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 4);
 
     /* space_style */
-    hbox = create_combo_widget (_("Space type: "),
-                                &__widget_space_type,
-                                (gpointer) &__config_space_type,
-                                (gpointer) &space_types);
+    widget = create_combo_widget (_("Space type: "),
+                                  &__widget_space_type,
+                                  (gpointer) &__config_space_type,
+                                  (gpointer) &space_types);
     gtk_widget_set_size_request (__widget_space_type, 100, -1);
-    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
+    gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 4);
 
     /* auto convert */
     __widget_auto_convert = gtk_check_button_new_with_mnemonic (_("Start conversion on inputting a comma or a period."));
@@ -562,6 +762,31 @@ create_options_page ()
     gtk_box_pack_start (GTK_BOX (vbox), __widget_auto_convert, FALSE, FALSE, 4);
     gtk_container_set_border_width (GTK_CONTAINER (__widget_auto_convert), 4);
 
+    /* close candidate window on select */
+    __widget_close_cand_win_on_select = gtk_check_button_new_with_mnemonic (_("Close candidate window when select a candidate directly."));
+    gtk_widget_show (__widget_close_cand_win_on_select);
+    gtk_box_pack_start (GTK_BOX (vbox), __widget_close_cand_win_on_select, FALSE, FALSE, 4);
+    gtk_container_set_border_width (GTK_CONTAINER (__widget_close_cand_win_on_select), 4);
+
+    // Connect all signals.
+    g_signal_connect ((gpointer) __widget_close_cand_win_on_select, "toggled",
+                      G_CALLBACK (on_default_toggle_button_toggled),
+                      &__config_close_cand_win_on_select);
+    g_signal_connect ((gpointer) __widget_auto_convert, "toggled",
+                      G_CALLBACK (on_default_toggle_button_toggled),
+                      &__config_auto_convert);
+
+    return vbox;
+}
+
+static GtkWidget *
+create_toolbar_page ()
+{
+    GtkWidget *vbox, *hbox, *label;
+
+    vbox = gtk_vbox_new (FALSE, 0);
+    gtk_widget_show (vbox);
+
     /* show/hide toolbar label */
     __widget_show_input_mode_label = gtk_check_button_new_with_mnemonic (_("Show _input mode label"));
     gtk_widget_show (__widget_show_input_mode_label);
@@ -578,10 +803,33 @@ create_options_page ()
     gtk_box_pack_start (GTK_BOX (vbox), __widget_show_period_style_label, FALSE, FALSE, 4);
     gtk_container_set_border_width (GTK_CONTAINER (__widget_show_period_style_label), 4);
 
+    /* dictionary menu */
+    __widget_show_dict_label = gtk_check_button_new_with_mnemonic (_("Show _dictionary menu label"));
+    gtk_widget_show (__widget_show_dict_label);
+    gtk_box_pack_start (GTK_BOX (vbox), __widget_show_dict_label, FALSE, FALSE, 4);
+    gtk_container_set_border_width (GTK_CONTAINER (__widget_show_dict_label), 4);
+
+    hbox = gtk_hbox_new (FALSE, 0);
+    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
+    gtk_widget_show (hbox);
+    label = gtk_label_new ("    ");
+    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
+    gtk_widget_show (label);
+    __widget_show_dict_admin_label = gtk_check_button_new_with_mnemonic (_("Show _edit dictionary label"));
+    gtk_widget_show (__widget_show_dict_admin_label);
+    gtk_box_pack_start (GTK_BOX (hbox), __widget_show_dict_admin_label, FALSE, FALSE, 0);
+
+    hbox = gtk_hbox_new (FALSE, 0);
+    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 4);
+    gtk_widget_show (hbox);
+    label = gtk_label_new ("    ");
+    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
+    gtk_widget_show (label);
+    __widget_show_add_word_label = gtk_check_button_new_with_mnemonic (_("Show _add word label"));
+    gtk_widget_show (__widget_show_add_word_label);
+    gtk_box_pack_start (GTK_BOX (hbox), __widget_show_add_word_label, FALSE, FALSE, 0);
+
     // Connect all signals.
-    g_signal_connect ((gpointer) __widget_auto_convert, "toggled",
-                      G_CALLBACK (on_default_toggle_button_toggled),
-                      &__config_auto_convert);
     g_signal_connect ((gpointer) __widget_show_input_mode_label, "toggled",
                       G_CALLBACK (on_default_toggle_button_toggled),
                       &__config_show_input_mode_label);
@@ -591,65 +839,106 @@ create_options_page ()
     g_signal_connect ((gpointer) __widget_show_period_style_label, "toggled",
                       G_CALLBACK (on_default_toggle_button_toggled),
                       &__config_show_period_style_label);
+    g_signal_connect ((gpointer) __widget_show_dict_label, "toggled",
+                      G_CALLBACK (on_default_toggle_button_toggled),
+                      &__config_show_dict_label);
+    g_signal_connect ((gpointer) __widget_show_dict_label, "toggled",
+                      G_CALLBACK (on_dict_menu_label_toggled),
+                      NULL);
+    g_signal_connect ((gpointer) __widget_show_dict_admin_label, "toggled",
+                      G_CALLBACK (on_default_toggle_button_toggled),
+                      &__config_show_dict_admin_label);
+    g_signal_connect ((gpointer) __widget_show_add_word_label, "toggled",
+                      G_CALLBACK (on_default_toggle_button_toggled),
+                      &__config_show_add_word_label);
 
     return vbox;
 }
 
+#define APPEND_ENTRY(text, widget, i) \
+{ \
+    label = gtk_label_new (NULL); \
+    gtk_label_set_text_with_mnemonic (GTK_LABEL (label), text); \
+    gtk_widget_show (label); \
+    gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5); \
+    gtk_misc_set_padding (GTK_MISC (label), 4, 0); \
+    gtk_table_attach (GTK_TABLE (table), label, 0, 1, i, i+1, \
+                      (GtkAttachOptions) (GTK_FILL), \
+                      (GtkAttachOptions) (GTK_FILL), 4, 4); \
+    widget = gtk_entry_new (); \
+    gtk_widget_show (widget); \
+    gtk_table_attach (GTK_TABLE (table), widget, 1, 2, i, i+1, \
+                      (GtkAttachOptions) (GTK_FILL|GTK_EXPAND), \
+                      (GtkAttachOptions) (GTK_FILL), 4, 4); \
+}
+
+static GtkWidget *
+create_dict_page (void)
+{
+    GtkWidget *table;
+    GtkWidget *label;
+
+    table = gtk_table_new (3, 3, FALSE);
+    gtk_widget_show (table);
+
+    // dict admin command
+    APPEND_ENTRY(_("Edit dictionary command:"), __widget_dict_admin_command, 0);
+
+    // add word command
+    APPEND_ENTRY(_("Add word command:"), __widget_add_word_command, 1);
+
+    // signals
+    g_signal_connect ((gpointer) __widget_dict_admin_command, "changed",
+                      G_CALLBACK (on_default_editable_changed),
+                      &__config_dict_admin_command);
+    g_signal_connect ((gpointer) __widget_add_word_command, "changed",
+                      G_CALLBACK (on_default_editable_changed),
+                      &__config_add_word_command);
+
+    return table;
+}
+
 static GtkWidget *
 create_keyboard_page (unsigned int page)
 {
     GtkWidget *table;
     GtkWidget *label;
-    unsigned int i, start = 0, end = 0;
 
-    if (page >= key_conf_pages)
+    if (page >= __key_conf_pages_num)
         return NULL;
 
-    for (i = 0; i < page; i++)
-        start += key_nums[i];
-    end = start + key_nums[page];
+    KeyboardConfigData *data = __key_conf_pages[page].data;
 
     table = gtk_table_new (3, 3, FALSE);
     gtk_widget_show (table);
 
     // Create keyboard setting.
-    for (i = start; i < end /*__config_keyboards [i].key*/; ++ i) {
-        label = gtk_label_new (NULL);
-        gtk_label_set_text_with_mnemonic (GTK_LABEL (label), _(__config_keyboards[i].label));
-        gtk_widget_show (label);
-        gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
-        gtk_misc_set_padding (GTK_MISC (label), 4, 0);
-        gtk_table_attach (GTK_TABLE (table), label, 0, 1, i, i+1,
+    for (unsigned int i = 0; data[i].key; ++ i) {
+        APPEND_ENTRY(_(data[i].label), data[i].entry, i);
+        gtk_entry_set_editable (GTK_ENTRY (data[i].entry), FALSE);
+
+        data[i].button = gtk_button_new_with_label ("...");
+        gtk_widget_show (data[i].button);
+        gtk_table_attach (GTK_TABLE (table), data[i].button, 2, 3, i, i+1,
                           (GtkAttachOptions) (GTK_FILL),
                           (GtkAttachOptions) (GTK_FILL), 4, 4);
-
-        __config_keyboards [i].entry = gtk_entry_new ();
-        gtk_widget_show (__config_keyboards [i].entry);
-        gtk_table_attach (GTK_TABLE (table), __config_keyboards [i].entry, 1, 2, i, i+1,
-                          (GtkAttachOptions) (GTK_FILL|GTK_EXPAND),
-                          (GtkAttachOptions) (GTK_FILL), 4, 4);
-        gtk_entry_set_editable (GTK_ENTRY (__config_keyboards[i].entry), FALSE);
-
-        __config_keyboards[i].button = gtk_button_new_with_label ("...");
-        gtk_widget_show (__config_keyboards[i].button);
-        gtk_table_attach (GTK_TABLE (table), __config_keyboards[i].button, 2, 3, i, i+1,
-                          (GtkAttachOptions) (GTK_FILL),
-                          (GtkAttachOptions) (GTK_FILL), 4, 4);
-        gtk_label_set_mnemonic_widget (GTK_LABEL (label), __config_keyboards[i].button);
+        gtk_label_set_mnemonic_widget (GTK_LABEL (label), data[i].button);
     }
 
-    for (i = start; i < end /*__config_keyboards [i].key*/; ++ i) {
-        g_signal_connect ((gpointer) __config_keyboards [i].button, "clicked",
+    for (unsigned int i = 0; data[i].key; ++ i) {
+        g_signal_connect ((gpointer) data[i].button, "clicked",
                           G_CALLBACK (on_default_key_selection_clicked),
-                          &(__config_keyboards [i]));
-        g_signal_connect ((gpointer) __config_keyboards [i].entry, "changed",
+                          &(data[i]));
+        g_signal_connect ((gpointer) data[i].entry, "changed",
                           G_CALLBACK (on_default_editable_changed),
-                          &(__config_keyboards [i].data));
+                          &(data[i].data));
     }
 
-    for (i = start; i < end /*__config_keyboards [i].key*/; ++ i) {
-        gtk_tooltips_set_tip (__widget_tooltips, __config_keyboards [i].entry,
-                              _(__config_keyboards [i].tooltip), NULL);
+    if (!__widget_tooltips)
+        __widget_tooltips = gtk_tooltips_new();
+    for (unsigned int i = 0; data[i].key; ++ i) {
+        gtk_tooltips_set_tip (__widget_tooltips, data[i].entry,
+                              _(data[i].tooltip), NULL);
     }
 
     return table;
@@ -662,6 +951,7 @@ create_setup_window ()
 
     if (!window) {
         GtkWidget *notebook = gtk_notebook_new();
+        gtk_notebook_popup_enable(GTK_NOTEBOOK(notebook));
         gtk_widget_show (notebook);
         window = notebook;
         gtk_notebook_set_scrollable (GTK_NOTEBOOK (notebook), TRUE);
@@ -672,23 +962,37 @@ create_setup_window ()
         gtk_widget_show (label);
         gtk_notebook_append_page (GTK_NOTEBOOK (notebook), page, label);
 
-        setup_widget_value ();
-
         // Create the second page.
-        for (unsigned int i = 0; i < key_conf_pages; i++) {
+        page = create_toolbar_page ();
+        label = gtk_label_new (_("Toolbar"));
+        gtk_widget_show (label);
+        gtk_notebook_append_page (GTK_NOTEBOOK (notebook), page, label);
+
+        // Create the third page.
+        page = create_dict_page ();
+        label = gtk_label_new (_("Dictionary"));
+        gtk_widget_show (label);
+        gtk_notebook_append_page (GTK_NOTEBOOK (notebook), page, label);
+
+        // Create the key bind pages.
+        for (unsigned int i = 0; i < __key_conf_pages_num; i++) {
             page = create_keyboard_page (i);
-            label = gtk_label_new (_(key_conf_page_labels[i]));
+            label = gtk_label_new (_(__key_conf_pages[i].label));
             gtk_widget_show (label);
             gtk_notebook_append_page (GTK_NOTEBOOK (notebook), page, label);
         }
 
+        // for preventing enabling left arrow.
+	gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), 1);
+        gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), 0);
+
         setup_widget_value ();
     }
 
     return window;
 }
 
-void
+static void
 setup_combo_value (GtkCombo *combo,
                    ComboConfigData *data, const String & str)
 {
@@ -708,7 +1012,7 @@ setup_combo_value (GtkCombo *combo,
         gtk_entry_set_text (GTK_ENTRY (combo->entry), defval);
 }
 
-void
+static void
 setup_widget_value ()
 {
     if (__widget_typing_method) {
@@ -732,6 +1036,12 @@ setup_widget_value ()
             __config_auto_convert);
     }
 
+    if (__widget_close_cand_win_on_select) {
+        gtk_toggle_button_set_active (
+            GTK_TOGGLE_BUTTON (__widget_close_cand_win_on_select),
+            __config_close_cand_win_on_select);
+    }
+
     if (__widget_show_input_mode_label) {
         gtk_toggle_button_set_active (
             GTK_TOGGLE_BUTTON (__widget_show_input_mode_label),
@@ -750,16 +1060,48 @@ setup_widget_value ()
             __config_show_period_style_label);
     }
 
-    for (int i = 0; __config_keyboards [i].key; ++ i) {
-        if (__config_keyboards [i].entry) {
-            gtk_entry_set_text (
-                GTK_ENTRY (__config_keyboards [i].entry),
-                __config_keyboards [i].data.c_str ());
+    if (__widget_dict_admin_command) {
+        gtk_entry_set_text (
+            GTK_ENTRY (__widget_dict_admin_command),
+            __config_dict_admin_command.c_str ());
+    }
+
+    if (__widget_add_word_command) {
+        gtk_entry_set_text (
+            GTK_ENTRY (__widget_add_word_command),
+            __config_add_word_command.c_str ());
+    }
+
+    if (__widget_show_dict_label) {
+        gtk_toggle_button_set_active (
+            GTK_TOGGLE_BUTTON (__widget_show_dict_label),
+            __config_show_dict_label);
+    }
+
+    if (__widget_show_dict_admin_label) {
+        gtk_toggle_button_set_active (
+            GTK_TOGGLE_BUTTON (__widget_show_dict_admin_label),
+            __config_show_dict_admin_label);
+    }
+
+    if (__widget_show_add_word_label) {
+        gtk_toggle_button_set_active (
+            GTK_TOGGLE_BUTTON (__widget_show_add_word_label),
+            __config_show_add_word_label);
+    }
+
+    for (unsigned int j = 0; j < __key_conf_pages_num; ++j) {
+        for (unsigned int i = 0; __key_conf_pages[j].data[i].key; ++ i) {
+            if (__key_conf_pages[j].data[i].entry) {
+                gtk_entry_set_text (
+                    GTK_ENTRY (__key_conf_pages[j].data[i].entry),
+                    __key_conf_pages[j].data[i].data.c_str ());
+            }
         }
     }
 }
 
-void
+static void
 load_config (const ConfigPointer &config)
 {
     if (!config.null ()) {
@@ -775,6 +1117,16 @@ load_config (const ConfigPointer &config)
         __config_auto_convert =
             config->read (String (SCIM_ANTHY_CONFIG_AUTO_CONVERT_ON_PERIOD),
                           __config_auto_convert);
+        __config_close_cand_win_on_select =
+            config->read (String (SCIM_ANTHY_CONFIG_CLOSE_CAND_WIN_ON_SELECT),
+                          __config_close_cand_win_on_select);
+        __config_dict_admin_command =
+            config->read (String (SCIM_ANTHY_CONFIG_DICT_ADMIN_COMMAND),
+                          __config_dict_admin_command);
+        __config_add_word_command =
+            config->read (String (SCIM_ANTHY_CONFIG_ADD_WORD_COMMAND),
+                          __config_add_word_command);
+
         __config_show_input_mode_label =
             config->read (String (SCIM_ANTHY_CONFIG_SHOW_INPUT_MODE_LABEL),
                           __config_show_input_mode_label);
@@ -784,11 +1136,22 @@ load_config (const ConfigPointer &config)
         __config_show_period_style_label =
             config->read (String (SCIM_ANTHY_CONFIG_SHOW_PERIOD_STYLE_LABEL),
                           __config_show_period_style_label);
+        __config_show_dict_label =
+            config->read (String (SCIM_ANTHY_CONFIG_SHOW_DICT_LABEL),
+                          __config_show_dict_label);
+        __config_show_dict_admin_label =
+            config->read (String (SCIM_ANTHY_CONFIG_SHOW_DICT_ADMIN_LABEL),
+                          __config_show_dict_admin_label);
+        __config_show_add_word_label =
+            config->read (String (SCIM_ANTHY_CONFIG_SHOW_ADD_WORD_LABEL),
+                          __config_show_add_word_label);
 
-        for (int i = 0; __config_keyboards [i].key; ++ i) {
-            __config_keyboards [i].data =
-                config->read (String (__config_keyboards [i].key),
-                              __config_keyboards [i].data);
+        for (unsigned int j = 0; j < __key_conf_pages_num; ++ j) {
+            for (unsigned int i = 0; __key_conf_pages[j].data[i].key; ++ i) {
+                __key_conf_pages[j].data[i].data =
+                    config->read (String (__key_conf_pages[j].data[i].key),
+                                  __key_conf_pages[j].data[i].data);
+            }
         }
 
         setup_widget_value ();
@@ -797,7 +1160,7 @@ load_config (const ConfigPointer &config)
     }
 }
 
-void
+static void
 save_config (const ConfigPointer &config)
 {
     if (!config.null ()) {
@@ -809,23 +1172,38 @@ save_config (const ConfigPointer &config)
                         __config_space_type);
         config->write (String (SCIM_ANTHY_CONFIG_AUTO_CONVERT_ON_PERIOD),
                         __config_auto_convert);
+        config->write (String (SCIM_ANTHY_CONFIG_CLOSE_CAND_WIN_ON_SELECT),
+                        __config_close_cand_win_on_select);
+        config->write (String (SCIM_ANTHY_CONFIG_DICT_ADMIN_COMMAND),
+                        __config_dict_admin_command);
+        config->write (String (SCIM_ANTHY_CONFIG_ADD_WORD_COMMAND),
+                        __config_add_word_command);
+
         config->write (String (SCIM_ANTHY_CONFIG_SHOW_INPUT_MODE_LABEL),
                         __config_show_input_mode_label);
         config->write (String (SCIM_ANTHY_CONFIG_SHOW_TYPING_METHOD_LABEL),
                         __config_show_typing_method_label);
         config->write (String (SCIM_ANTHY_CONFIG_SHOW_PERIOD_STYLE_LABEL),
                         __config_show_period_style_label);
+        config->write (String (SCIM_ANTHY_CONFIG_SHOW_DICT_LABEL),
+                        __config_show_dict_label);
+        config->write (String (SCIM_ANTHY_CONFIG_SHOW_DICT_ADMIN_LABEL),
+                        __config_show_dict_admin_label);
+        config->write (String (SCIM_ANTHY_CONFIG_SHOW_ADD_WORD_LABEL),
+                        __config_show_add_word_label);
 
-        for (int i = 0; __config_keyboards [i].key; ++ i) {
-            config->write (String (__config_keyboards [i].key),
-                          __config_keyboards [i].data);
+        for (unsigned int j = 0; j < __key_conf_pages_num; j++) {
+            for (unsigned int i = 0; __key_conf_pages[j].data[i].key; ++ i) {
+                config->write (String (__key_conf_pages[j].data[i].key),
+                               __key_conf_pages[j].data[i].data);
+            }
         }
 
         __have_changed = false;
     }
 }
 
-bool
+static bool
 query_changed ()
 {
     return __have_changed;
@@ -908,3 +1286,17 @@ on_default_combo_changed (GtkEditable *editable,
         }
     }
 }
+
+static void
+on_dict_menu_label_toggled (GtkToggleButton *togglebutton,
+                            gpointer         user_data)
+{
+    bool active = gtk_toggle_button_get_active (togglebutton);
+    if (__widget_show_dict_admin_label)
+        gtk_widget_set_sensitive (__widget_show_dict_admin_label, active);
+    if (__widget_show_add_word_label)
+        gtk_widget_set_sensitive (__widget_show_add_word_label, active);
+}
+/*
+vi:ts=4:nowrap:ai:expandtab
+*/
